import abc
import ast
from dataclasses import dataclass, field
from typing import Optional, List

from transpiler.phases.typing.scope import Scope
from transpiler.phases.typing.types import BaseType, TY_NONE, TypeType, TY_SELF, TypeVariable
from transpiler.phases.utils import NodeVisitorSeq


@dataclass
class TypeAnnotationVisitor(NodeVisitorSeq):
    scope: Scope
    cur_class: Optional[TypeType] = None
    typevars: List[TypeVariable] = field(default_factory=list)

    def visit_str(self, node: str) -> BaseType:
        if node in ("Self", "self") and self.cur_class:
            if isinstance(self.cur_class.type_object, abc.ABCMeta) or self.cur_class.type_object.is_protocol_gen or self.cur_class.type_object.is_protocol:
                return TY_SELF
            else:
                return self.cur_class.type_object
        if existing := self.scope.get(node):
            ty = existing.type
            if isinstance(ty, TypeType) and isinstance(ty.type_object, TypeVariable):
                if existing is not self.scope.vars.get(node, None):
                    # Type variable from outer scope, so we copy it
                    ty = TypeVariable(ty.type_object.name)
                    self.scope.declare_local(node, TypeType(ty)) # todo: unneeded?
                    self.typevars.append(ty)
            if isinstance(ty, TypeType):
                return ty.type_object
            return ty

        raise NameError(node)

    def visit_Name(self, node: ast.Name) -> BaseType:
        return self.visit_str(node.id)

    def visit_Constant(self, node: ast.Constant) -> BaseType:
        if node.value is None:
            return TY_NONE
        if type(node.value) == str:
            return node.value
        raise NotImplementedError

    def visit_Subscript(self, node: ast.Subscript) -> BaseType:
        ty_op = self.visit(node.value)
        args = list(node.slice.elts) if type(node.slice) == ast.Tuple else [node.slice]
        args = [self.visit(arg) for arg in args]
        return ty_op(*args)
        # return TypeOperator([self.visit(node.value)], self.visit(node.slice.value))

    def visit_List(self, node: ast.List) -> List[BaseType]:
        return [self.visit(elt) for elt in node.elts]

    def visit_Attribute(self, node: ast.Attribute) -> BaseType:
        left = self.visit(node.value)
        res = left.members[node.attr]
        assert isinstance(res, TypeType)
        return res.type_object
        raise NotImplementedError(ast.unparse(node))
