import ast
import dataclasses
from dataclasses import dataclass

from transpiler.phases.typing.common import ScoperVisitor
from transpiler.phases.typing.expr import ScoperExprVisitor
from transpiler.phases.typing.class_ import ScoperClassVisitor
from transpiler.phases.typing.scope import VarDecl, VarKind, ScopeKind
from transpiler.phases.typing.types import BaseType, TypeVariable, FunctionType, IncompatibleTypesError, \
    Promise, TY_NONE, PromiseKind, TupleType, UserType, TypeType, ModuleType
from transpiler.phases.utils import PlainBlock


@dataclass
class ScoperBlockVisitor(ScoperVisitor):
    stdlib: bool = False

    def expr(self) -> ScoperExprVisitor:
        return ScoperExprVisitor(self.scope, self.root_decls)

    def visit_Pass(self, node: ast.Pass):
        pass

    def visit_Import(self, node: ast.Import):
        for alias in node.names:
            mod = self.scope.get(alias.name, VarKind.MODULE)
            if mod is None:
                raise NameError(alias.name)
            assert isinstance(mod, VarDecl), mod
            self.scope.vars[alias.asname or alias.name] = dataclasses.replace(mod, kind=VarKind.LOCAL)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module == "typing":
            return
        module = self.scope.get(node.module, VarKind.MODULE)
        if not module:
            raise NameError(node.module)
        if not isinstance(module.type, ModuleType):
            raise IncompatibleTypesError(f"{node.module} is not a module")
        for alias in node.names:
            thing = module.val.get(alias.name)
            if not thing:
                raise NameError(alias.name)
            self.scope.vars[alias.asname or alias.name] = VarDecl(VarKind.LOCAL, thing)

    def visit_Module(self, node: ast.Module):
        self.visit_block(node.body)

    def get_type(self, node: ast.expr) -> BaseType:
        if type := getattr(node, "type", None):
            return type
        self.expr().visit(node)
        return node.type
        # ntype = TypeVariable()
        # node.type = ntype
        # return ntype

    def visit_Assign(self, node: ast.Assign):
        if len(node.targets) != 1:
            raise NotImplementedError(node)
        target = node.targets[0]
        ty = self.get_type(node.value)
        try:
            node.is_declare = self.visit_assign_target(target, ty)
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"`{ast.unparse(node)}: {e}") from e

    def visit_AnnAssign(self, node: ast.AnnAssign):
        if node.value is not None:
            raise NotImplementedError(node)
        if node.simple != 1:
            raise NotImplementedError(node)
        if not isinstance(node.target, ast.Name):
            raise NotImplementedError(node)
        ty = self.visit_annotation(node.annotation)
        try:
            node.is_declare = self.visit_assign_target(node.target, ty)
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"`{ast.unparse(node)}: {e}")

    def visit_assign_target(self, target, decl_val: BaseType) -> bool:
        if isinstance(target, ast.Name):
            if target.id == "_":
                return False
            target.type = decl_val
            if vdecl := self.scope.get(target.id):
                vdecl.type.unify(decl_val)
                return False
            else:
                self.scope.vars[target.id] = VarDecl(VarKind.LOCAL, decl_val)
                if self.scope.kind == ScopeKind.FUNCTION_INNER:
                    self.root_decls[target.id] = VarDecl(VarKind.OUTER_DECL, decl_val)
                return True
        elif isinstance(target, ast.Tuple):
            if not (isinstance(decl_val, TupleType) and len(target.elts) == len(decl_val.args)):
                raise IncompatibleTypesError(f"Cannot unpack {decl_val} into {target}")
            decls = [self.visit_assign_target(t, ty) for t, ty in zip(target.elts, decl_val.args)]  # eager evaluated
            return any(decls)
        elif isinstance(target, ast.Attribute):
            attr_type = self.expr().visit(target)
            attr_type.unify(decl_val)
            return False
        elif isinstance(target, ast.Subscript):
            expr = self.expr()
            left = expr.visit(target.value)
            args = target.slice if type(target.slice) == tuple else [target.slice]
            args = [expr.visit(e) for e in args]
            if len(args) == 1:
                args = args[0]
            expr.make_dunder([left, args, decl_val], "setitem")
            return False
        else:
            raise NotImplementedError(ast.unparse(target))

    def visit_FunctionDef(self, node: ast.FunctionDef):
        argtypes = [self.visit_annotation(arg.annotation) for arg in node.args.args]
        rtype = Promise(self.visit_annotation(node.returns), PromiseKind.TASK)
        ftype = FunctionType(argtypes, rtype)
        self.scope.vars[node.name] = VarDecl(VarKind.LOCAL, ftype)
        scope = self.scope.child(ScopeKind.FUNCTION)
        scope.obj_type = ftype
        scope.function = scope
        node.inner_scope = scope
        node.type = ftype
        ftype.optional_at = 1 + len(node.args.args) - len(node.args.defaults)
        for arg, ty in zip(node.args.args, argtypes):
            scope.vars[arg.arg] = VarDecl(VarKind.LOCAL, ty)
        self.fdecls.append((node, rtype.return_type))

    def visit_ClassDef(self, node: ast.ClassDef):
        ctype = UserType(node.name)
        cttype = TypeType(ctype)
        self.scope.vars[node.name] = VarDecl(VarKind.LOCAL, cttype)
        scope = self.scope.child(ScopeKind.CLASS)
        scope.obj_type = ctype
        scope.class_ = scope
        node.inner_scope = scope
        node.type = ctype
        visitor = ScoperClassVisitor(scope, cur_class=cttype)
        visitor.visit_block(node.body)

    def visit_If(self, node: ast.If):
        scope = self.scope.child(ScopeKind.FUNCTION_INNER)
        node.inner_scope = scope
        self.expr().visit(node.test)
        then_scope = scope.child(ScopeKind.FUNCTION_INNER)
        then_visitor = ScoperBlockVisitor(then_scope, self.root_decls)
        then_visitor.visit_block(node.body)
        if node.orelse:
            else_scope = scope.child(ScopeKind.FUNCTION_INNER)
            else_visitor = ScoperBlockVisitor(else_scope, self.root_decls)
            else_visitor.visit_block(node.orelse)
            node.orelse_scope = else_scope

    def visit_While(self, node: ast.While):
        scope = self.scope.child(ScopeKind.FUNCTION_INNER)
        node.inner_scope = scope
        self.expr().visit(node.test)
        body_scope = scope.child(ScopeKind.FUNCTION_INNER)
        body_visitor = ScoperBlockVisitor(body_scope, self.root_decls)
        body_visitor.visit_block(node.body)
        if node.orelse:
            raise NotImplementedError(node.orelse)

    def visit_PlainBlock(self, node: PlainBlock):
        scope = self.scope.child(ScopeKind.FUNCTION_INNER)
        node.inner_scope = scope
        body_visitor = ScoperBlockVisitor(scope, self.root_decls)
        body_visitor.visit_block(node.body)

    def visit_For(self, node: ast.For):
        scope = self.scope.child(ScopeKind.FUNCTION_INNER)
        node.inner_scope = scope
        assert isinstance(node.target, ast.Name)
        var_var = TypeVariable()
        scope.vars[node.target.id] = VarDecl(VarKind.LOCAL, var_var)
        seq_type = self.expr().visit(node.iter)
        try:
            iter_type = seq_type.methods["__iter__"].return_type
        except:
            raise IncompatibleTypesError(f"{seq_type} is not iterable in `{ast.unparse(node.iter)}`")
        try:
            next_type = iter_type.methods["__next__"].return_type
        except:
            raise IncompatibleTypesError(f"iter({iter_type}) is not an iterator")
        var_var.unify(next_type)
        body_scope = scope.child(ScopeKind.FUNCTION_INNER)
        body_visitor = ScoperBlockVisitor(body_scope, self.root_decls)
        body_visitor.visit_block(node.body)
        if node.orelse:
            raise NotImplementedError(node.orelse)

    def visit_Expr(self, node: ast.Expr):
        self.expr().visit(node.value)

    def visit_Return(self, node: ast.Return):
        fct = self.scope.function
        if fct is None:
            raise IncompatibleTypesError("Return outside function")
        ftype = fct.obj_type
        assert isinstance(ftype, FunctionType)
        vtype = self.expr().visit(node.value) if node.value else TY_NONE
        vtype.unify(ftype.return_type.return_type if isinstance(ftype.return_type, Promise) else ftype.return_type)
        fct.has_return = True

    def visit_Global(self, node: ast.Global):
        for name in node.names:
            self.scope.function.vars[name] = VarDecl(VarKind.GLOBAL, None)
            if name not in self.scope.global_scope.vars:
                self.scope.global_scope.vars[name] = VarDecl(VarKind.LOCAL, None)

    def visit_AugAssign(self, node: ast.AugAssign):
        equivalent = ast.Assign(targets=[node.target], value=ast.BinOp(left=node.target, op=node.op, right=node.value))
        self.visit(equivalent)

    def visit(self, node: ast.AST):
        if isinstance(node, ast.AST):
            super().visit(node)
            node.scope = self.scope
        else:
            raise NotImplementedError(node)

    def visit_Break(self, node: ast.Break):
        pass  # TODO: check in loop

    def visit_Try(self, node: ast.Try):
        scope = self.scope.child(ScopeKind.FUNCTION_INNER)
        node.inner_scope = scope
        body_scope = scope.child(ScopeKind.FUNCTION_INNER)
        body_visitor = ScoperBlockVisitor(body_scope, self.root_decls)
        body_visitor.visit_block(node.body)
        # todo
        for handler in node.handlers:
            handler_scope = scope.child(ScopeKind.FUNCTION_INNER)
            handler_visitor = ScoperBlockVisitor(handler_scope, self.root_decls)
            handler_visitor.visit_block(handler.body)
        if node.orelse:
            else_scope = scope.child(ScopeKind.FUNCTION_INNER)
            else_visitor = ScoperBlockVisitor(else_scope, self.root_decls)
            else_visitor.visit_block(node.orelse)
        if node.finalbody:
            raise NotImplementedError(node.finalbody)