import abc
import ast
import inspect
from typing import List

from transpiler.phases.typing import ScopeKind, VarDecl, VarKind
from transpiler.phases.typing.common import ScoperVisitor
from transpiler.phases.typing.types import IncompatibleTypesError, BaseType, TupleType, TY_STR, TY_BOOL, TY_INT, \
    TY_COMPLEX, TY_NONE, FunctionType, PyList, TypeVariable, PySet, TypeType, PyDict, Promise, PromiseKind, UserType

DUNDER = {
    ast.Eq: "eq",
    ast.NotEq: "ne",
    ast.Mult: "mul",
    ast.Add: "add",
    ast.Sub: "sub",
    ast.Div: "truediv",
    ast.FloorDiv: "floordiv",
    ast.Mod: "mod",
    ast.Lt: "lt",
    ast.Gt: "gt",
    ast.GtE: "ge",
    ast.LtE: "le",
    ast.LShift: "lshift",
    ast.RShift: "rshift",
    ast.BitXor: "xor",
    ast.BitOr: "or",
    ast.BitAnd: "and",
    ast.USub: "neg",
    ast.UAdd: "pos",
    ast.Invert: "invert",
}

class ScoperExprVisitor(ScoperVisitor):
    def visit(self, node) -> BaseType:
        if existing := getattr(node, "type", None):
            return existing
        res = super().visit(node)
        if not res:
            raise NotImplementedError(f"`{ast.unparse(node)}` {type(node)}")
        res = res.resolve()
        node.type = res
        return res

    def visit_Tuple(self, node: ast.Tuple) -> BaseType:
        return TupleType(*[self.visit(e) for e in node.elts])

    def visit_Yield(self, node: ast.Yield) -> BaseType:
        ytype = self.visit(node.value)

        ftype = self.scope.function.obj_type.return_type
        assert isinstance(ftype, Promise)
        assert ftype.kind == PromiseKind.TASK
        ftype.kind = PromiseKind.GENERATOR
        ftype.return_type.unify(ytype)
        self.scope.function.has_return = True

        return TY_NONE

    def visit_Constant(self, node: ast.Constant) -> BaseType:
        if isinstance(node.value, str):
            return TY_STR
        elif isinstance(node.value, bool):
            return TY_BOOL
        elif isinstance(node.value, int):
            return TY_INT
        elif isinstance(node.value, complex):
            return TY_COMPLEX
        elif node.value is None:
            return TY_NONE
        else:
            raise NotImplementedError(node, type(node))

    def visit_Name(self, node: ast.Name) -> BaseType:
        obj = self.scope.get(node.id)
        if not obj:
            raise NameError(f"Name {node.id} is not defined")
        if isinstance(obj.type, TypeType) and isinstance(obj.type.type_object, TypeVariable):
            raise NameError(f"Use of type variable")
        return obj.type

    def visit_Compare(self, node: ast.Compare) -> BaseType:
        # todo:
        self.visit(node.left)
        for op, right in zip(node.ops, node.comparators):
            self.visit(right)
        return TY_BOOL
        #raise NotImplementedError(node)

    def visit_Call(self, node: ast.Call) -> BaseType:
        ftype = self.visit(node.func)
        if ftype.typevars:
            ftype = ftype.gen_sub(None, {v.name: TypeVariable(v.name) for v in ftype.typevars})
        try:
            rtype = self.visit_function_call(ftype, [self.visit(arg) for arg in node.args])
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"`{ast.unparse(node)}`: {e}")
        actual = rtype
        node.is_await = False
        if isinstance(actual, Promise) and actual.kind != PromiseKind.GENERATOR:
            node.is_await = True
            actual = actual.return_type.resolve()

            if self.scope.function and isinstance(actual, Promise) and actual.kind == PromiseKind.FORKED \
                    and isinstance(fty := self.scope.function.obj_type.return_type, Promise):
                fty.kind = PromiseKind.JOIN

        return actual

    def visit_function_call(self, ftype: BaseType, arguments: List[BaseType]):
        if isinstance(ftype, TypeType):# and isinstance(ftype.type_object, UserType):
            init: FunctionType = self.visit_getattr(ftype, "__init__").remove_self()
            init.return_type = ftype.type_object
            return self.visit_function_call(init, arguments)
        if not isinstance(ftype, FunctionType):
            raise IncompatibleTypesError(f"Cannot call {ftype}")
        #is_generic = any(isinstance(arg, TypeVariable) for arg in ftype.to_list())
        equivalent = FunctionType(arguments, ftype.return_type)
        try:
            ftype.unify(equivalent)
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"Cannot call {ftype} with ({(', '.join(map(str, arguments)))}): {e}")
        return ftype.return_type

    def visit_Lambda(self, node: ast.Lambda) -> BaseType:
        argtypes = [TypeVariable() for _ in node.args.args]
        rtype = TypeVariable()
        ftype = FunctionType(argtypes, rtype)
        scope = self.scope.child(ScopeKind.FUNCTION)
        scope.obj_type = ftype
        scope.function = scope
        node.inner_scope = scope
        node.body.scope = scope
        for arg, ty in zip(node.args.args, argtypes):
            scope.vars[arg.arg] = VarDecl(VarKind.LOCAL, ty)
        decls = {}
        visitor = ScoperExprVisitor(scope, decls)
        rtype.unify(visitor.visit(node.body))
        node.body.decls = decls
        return ftype

    def visit_BinOp(self, node: ast.BinOp) -> BaseType:
        left, right = map(self.visit, (node.left, node.right))
        try:
            return self.make_dunder([left, right], DUNDER[type(node.op)])
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"{e} in `{ast.unparse(node)}`")

    def visit_Attribute(self, node: ast.Attribute) -> BaseType:
        try:
            ltype = self.visit(node.value)
            return self.visit_getattr(ltype, node.attr)
        except Exception as e:
            raise IncompatibleTypesError(f"{e} in `{ast.unparse(node)}`")

    def visit_getattr(self, ltype: BaseType, name: str):
        bound = True
        if isinstance(ltype, TypeType):
            ltype = ltype.type_object
            bound = False
        if isinstance(ltype, abc.ABCMeta):
            ctor = ltype.__init__
            args = list(inspect.signature(ctor).parameters.values())[1:]
            if not all(arg.annotation == BaseType for arg in args):
                raise NotImplementedError("I don't know how to handle this type")
            ltype = ltype(*(TypeVariable() for _ in args))
        if attr := ltype.members.get(name):
            return attr
        if meth := ltype.methods.get(name):
            if bound:
                return meth.remove_self()
            else:
                return meth
        raise IncompatibleTypesError(f"Type {ltype} has no attribute {name}")

    def visit_List(self, node: ast.List) -> BaseType:
        if not node.elts:
            return PyList(TypeVariable())
        elems = [self.visit(e) for e in node.elts]
        if len(set(elems)) != 1:
            raise NotImplementedError("List with different types not handled yet")
        return PyList(elems[0])

    def visit_Set(self, node: ast.Set) -> BaseType:
        if not node.elts:
            return PySet(TypeVariable())
        elems = [self.visit(e) for e in node.elts]
        if len(set(elems)) != 1:
            raise NotImplementedError("Set with different types not handled yet")
        return PySet(elems[0])

    def visit_Dict(self, node: ast.Dict) -> BaseType:
        if not node.keys:
            return PyDict(TypeVariable(), TypeVariable())
        keys = [self.visit(e) for e in node.keys]
        values = [self.visit(e) for e in node.values]
        if len(set(keys)) != 1 or len(set(values)) != 1:
            raise NotImplementedError(f"Dict with different types not handled yet in `{ast.unparse(node)}`")
        return PyDict(keys[0], values[0])

    def visit_Subscript(self, node: ast.Subscript) -> BaseType:
        left = self.visit(node.value)
        args = node.slice if type(node.slice) == tuple else [node.slice]
        args = [self.visit(e) for e in args]
        if isinstance(left, TypeType) and isinstance(left.type_object, abc.ABCMeta):
            # generic
            return TypeType(left.type_object(*[arg.type_object if isinstance(arg, TypeType) else arg for arg in args]))
            pass
        return self.make_dunder([left, *args], "getitem")

    def visit_UnaryOp(self, node: ast.UnaryOp) -> BaseType:
        val = self.visit(node.operand)
        if isinstance(node.op, ast.Not):
            return TY_BOOL
        try:
            return self.make_dunder([val], DUNDER[type(node.op)])
        except IncompatibleTypesError as e:
            raise IncompatibleTypesError(f"{e} in `{ast.unparse(node)}`")

    def visit_IfExp(self, node: ast.IfExp) -> BaseType:
        self.visit(node.test)
        then = self.visit(node.body)
        else_ = self.visit(node.orelse)
        if then != else_:
            raise NotImplementedError("IfExp with different types not handled yet")
        return then

    def make_dunder(self, args: List[BaseType], name: str) -> BaseType:
        return self.visit_function_call(
            self.visit_getattr(TypeType(args[0]), f"__{name}__"),
            args
        )