import ast
import dataclasses
from abc import ABCMeta
from dataclasses import dataclass, field
from typing import Optional, List, Dict

from transpiler.phases.typing.annotations import TypeAnnotationVisitor
from transpiler.phases.typing.common import PRELUDE
from transpiler.phases.typing.expr import ScoperExprVisitor
from transpiler.phases.typing.scope import Scope, VarDecl, VarKind, ScopeKind
from transpiler.phases.typing.types import BaseType, TypeOperator, FunctionType, TY_VARARG, TypeType, TypeVariable
from transpiler.phases.utils import NodeVisitorSeq



@dataclass
class StdlibVisitor(NodeVisitorSeq):
    scope: Scope = field(default_factory=lambda: PRELUDE)
    cur_class: Optional[BaseType] = None
    typevars: Dict[str, BaseType] = field(default_factory=dict)

    def expr(self) -> ScoperExprVisitor:
        return ScoperExprVisitor(self.scope)

    def visit_Module(self, node: ast.Module):
        for stmt in node.body:
            self.visit(stmt)

    def visit_Assign(self, node: ast.Assign):
        self.scope.vars[node.targets[0].id] = VarDecl(VarKind.LOCAL, self.visit(node.value))

    def visit_AnnAssign(self, node: ast.AnnAssign):
        ty = self.anno().visit(node.annotation)
        if self.cur_class:
            assert isinstance(self.cur_class, TypeType)
            if isinstance(self.cur_class.type_object, ABCMeta):
                raise NotImplementedError
            else:
                self.cur_class.type_object.members[node.target.id] = ty.gen_sub(self.cur_class.type_object, self.typevars)
        self.scope.vars[node.target.id] = VarDecl(VarKind.LOCAL, ty)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        pass

    def visit_Import(self, node: ast.Import):
        pass

    def visit_ClassDef(self, node: ast.ClassDef):
        if existing := self.scope.get(node.name):
            ty = existing.type
        else:
            class TheType(TypeOperator):
                def __init__(self, *args):
                    super().__init__(args, node.name)
            ty = TypeType(TheType)
            self.scope.vars[node.name] = VarDecl(VarKind.LOCAL, ty)
        typevars = []
        for b in node.bases:
            if isinstance(b, ast.Subscript):
                if isinstance(b.slice, ast.Name):
                    sliceval = [b.slice.id]
                elif isinstance(b.slice, ast.Tuple):
                    sliceval = [n.id for n in b.slice.elts]
                if isinstance(b.value, ast.Name) and b.value.id == "Generic":
                    typevars = sliceval
                elif isinstance(b.value, ast.Name) and b.value.id == "Protocol":
                    typevars = sliceval
                    ty.type_object.is_protocol_gen = True
                else:
                    idxs = [typevars.index(v) for v in sliceval]
                    parent = self.visit(b.value)
                    assert isinstance(parent, TypeType)
                    assert isinstance(ty.type_object, ABCMeta)
                    ty.type_object.gen_parents.append(lambda selfvars: parent.type_object(*[selfvars[i] for i in idxs]))
            else:
                if isinstance(b, ast.Name) and b.id == "Protocol":
                    ty.type_object.is_protocol_gen = True
                else:
                    parent = self.visit(b)
                    assert isinstance(parent, TypeType)
                    if isinstance(ty.type_object, ABCMeta):
                        ty.type_object.gen_parents.append(parent.type_object)
                    else:
                        ty.type_object.parents.append(parent.type_object)
        if not typevars and not existing:
            ty.type_object = ty.type_object()
        cl_scope = self.scope.child(ScopeKind.CLASS)
        visitor = StdlibVisitor(cl_scope, ty)
        for var in typevars:
            visitor.typevars[var] = TypeType(TypeVariable(var))
        for stmt in node.body:
            visitor.visit(stmt)

    def visit_Pass(self, node: ast.Pass):
        pass

    def visit_FunctionDef(self, node: ast.FunctionDef):
        tc = node.type_comment # todo : lire les commetnaries de type pour les fonctions génériques sinon trouver autre chose
        arg_visitor = TypeAnnotationVisitor(self.scope.child(ScopeKind.FUNCTION), self.cur_class)
        arg_types = [arg_visitor.visit(arg.annotation or arg.arg) for arg in node.args.args]
        ret_type = arg_visitor.visit(node.returns)
        ty = FunctionType(arg_types, ret_type)
        ty.typevars = arg_visitor.typevars
        if node.args.vararg:
            ty.variadic = True
        ty.optional_at = 1 + len(node.args.args) - len(node.args.defaults)
        if self.cur_class:
            assert isinstance(self.cur_class, TypeType)
            if isinstance(self.cur_class.type_object, ABCMeta):
                self.cur_class.type_object.gen_methods[node.name] = lambda t: ty.gen_sub(t, self.typevars)
            else:
                self.cur_class.type_object.methods[node.name] = ty.gen_sub(self.cur_class.type_object, self.typevars)
        self.scope.vars[node.name] = VarDecl(VarKind.LOCAL, ty)

    def visit_Assert(self, node: ast.Assert):
        if isinstance(node.test, ast.UnaryOp) and isinstance(node.test.op, ast.Not):
            oper = node.test.operand
            try:
                res = self.expr().visit(oper)
            except:
                print("Type of", ast.unparse(oper), ":=", "INVALID")
            else:
                raise AssertionError(f"Assertion should fail, got {res} for {ast.unparse(oper)}")
        else:
            print("Type of", ast.unparse(node.test), ":=", self.expr().visit(node.test))

    def visit_Call(self, node: ast.Call) -> BaseType:
        ty_op = self.visit(node.func)
        if isinstance(ty_op, TypeType):
            return TypeType(ty_op.type_object(*[ast.literal_eval(arg) for arg in node.args]))
        raise NotImplementedError(ast.unparse(node))

    def anno(self) -> "TypeAnnotationVisitor":
        return TypeAnnotationVisitor(self.scope, self.cur_class)

    def visit_str(self, node: str) -> BaseType:
        if existing := self.scope.get(node):
            return existing.type
        raise NameError(node)

    def visit_Name(self, node: ast.Name) -> BaseType:
        return self.visit_str(node.id)