import dataclasses
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from itertools import zip_longest
from typing import Dict, Optional, List, ClassVar, Callable


class IncompatibleTypesError(Exception):
    pass


@dataclass(eq=False)
class BaseType(ABC):
    members: Dict[str, "BaseType"] = field(default_factory=dict, init=False)
    methods: Dict[str, "FunctionType"] = field(default_factory=dict, init=False)
    parents: List["BaseType"] = field(default_factory=list, init=False)
    typevars: List["TypeVariable"] = field(default_factory=list, init=False)


    def get_parents(self) -> List["BaseType"]:
        return self.parents

    def resolve(self) -> "BaseType":
        return self

    @abstractmethod
    def unify_internal(self, other: "BaseType"):
        pass

    def unify(self, other: "BaseType"):
        a, b = self.resolve(), other.resolve()
        if isinstance(b, TypeVariable):
            a, b = b, a
        a.unify_internal(b)

    def contains(self, other: "BaseType") -> bool:
        needle, haystack = other.resolve(), self.resolve()
        return (needle is haystack) or haystack.contains_internal(needle)

    @abstractmethod
    def contains_internal(self, other: "BaseType") -> bool:
        pass

    # @abstractmethod
    # def clone(self) -> "BaseType":
    #     pass

    def gen_sub(self, this: "BaseType", typevars: Dict[str, "BaseType"], cache=None) -> "Self":
        return self

    def to_list(self) -> List["BaseType"]:
        return [self]


T = typing.TypeVar("T")


class MagicType(BaseType, typing.Generic[T]):
    val: T

    def __init__(self, val: T):
        super().__init__()
        self.val = val

    def unify_internal(self, other: "BaseType"):
        if type(self) != type(other) or self.val != other.val:
            raise IncompatibleTypesError()

    def contains_internal(self, other: "BaseType") -> bool:
        return False

    def __str__(self):
        return str(self.val)

    def clone(self) -> "BaseType":
        return MagicType(self.val)


cur_var = 0


@dataclass(eq=False)
class TypeVariable(BaseType):
    name: str = field(default_factory=lambda: chr(ord('a') + cur_var))
    resolved: Optional[BaseType] = None

    def __str__(self):
        if self.resolved is None:
            return self.name
        return str(self.resolved)

    def resolve(self) -> BaseType:
        if self.resolved is None:
            return self
        return self.resolved.resolve()

    def unify_internal(self, other: BaseType):
        if self is not other:
            if other.contains(self):
                raise ValueError(f"Recursive type: {self} and {other}")
            self.resolved = other

    def contains_internal(self, other: BaseType) -> bool:
        return self.resolve() is other.resolve()

    def gen_sub(self, this: "BaseType", typevars, cache=None) -> "Self":
        if match := typevars.get(self.name):
            return match
        return self


GenMethodFactory = Callable[["BaseType"], "FunctionType"]


@dataclass(eq=False)
class TypeOperator(BaseType, ABC):
    args: List[BaseType]
    name: str = None
    variadic: bool = False
    optional_at: Optional[int] = None
    gen_methods: ClassVar[Dict[str, GenMethodFactory]] = {}
    gen_parents: ClassVar[List[BaseType]] = []
    is_protocol: bool = False
    is_protocol_gen: ClassVar[bool] = False
    match_cache: set["TypeOperator"] = field(default_factory=set, init=False)
    is_reference: bool = False

    @staticmethod
    def make_type(name: str):
        class TheType(TypeOperator):
            def __init__(self):
                super().__init__([], name)
        return TheType()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.gen_methods = {}
        cls.gen_parents = []

    def __post_init__(self):
        if self.name is None:
            self.name = self.__class__.__name__
        for name, factory in self.gen_methods.items():
            self.methods[name] = factory(self)
        for gp in self.gen_parents:
            if not isinstance(gp, BaseType):
                gp = gp(self.args)
            self.parents.append(gp)
            self.methods = {**gp.methods, **self.methods}
        self.is_protocol = self.is_protocol or self.is_protocol_gen

    def matches_protocol(self, protocol: "TypeOperator"):
        if hash(protocol) in self.match_cache:
            return
        try:
            dupl = protocol.gen_sub(self, {v.name: (TypeVariable(v.name) if isinstance(v.resolve(), TypeVariable) else v) for v in protocol.args})
            self.match_cache.add(hash(protocol))
            for name, ty in dupl.methods.items():
                corresp = self.methods[name]
                corresp.remove_self().unify(ty.remove_self())
        except Exception as e:
            self.match_cache.remove(hash(protocol))
            raise IncompatibleTypesError(f"Type {self} doesn't implement protocol {protocol}: {e}")

    def unify_internal(self, other: BaseType):
        if not isinstance(other, TypeOperator):
            raise IncompatibleTypesError()
        if other.is_protocol and not self.is_protocol:
            return other.unify_internal(self)
        if self.is_protocol and not other.is_protocol:
            return other.matches_protocol(self)
        if len(self.args) < len(other.args):
            return other.unify_internal(self)
        assert self.is_protocol == other.is_protocol
        if type(self) != type(other):
            for parent in other.get_parents():
                try:
                    self.unify(parent)
                except IncompatibleTypesError:
                    pass
                else:
                    return
            for parent in self.get_parents():
                try:
                    parent.unify(other)
                except IncompatibleTypesError:
                    pass
                else:
                    return
            raise IncompatibleTypesError(f"Cannot unify {self} and {other} with different type and no common parents")
        if len(self.args) == 0:
            if self.name != other.name:
                raise IncompatibleTypesError(f"Cannot unify {self} and {other}")
        for i, (a, b) in enumerate(zip_longest(self.args, other.args)):
            if a is None and self.variadic or b is None and other.variadic:
                continue

            if a is not None and b is None:
                if self.optional_at is not None and i >= self.optional_at:
                    continue
                else:
                    raise IncompatibleTypesError(f"Cannot unify {self} and {other}, not enough arguments")

            if isinstance(a, BaseType) and isinstance(b, BaseType):
                a.unify(b)
            else:
                if a != b:
                    raise IncompatibleTypesError(f"Cannot unify {a} and {b}")

    def contains_internal(self, other: "BaseType") -> bool:
        return any(arg.contains(other) for arg in self.args)

    def __str__(self):
        return self.name + (f"<{', '.join(map(str, self.args))}>" if self.args else "")

    def __hash__(self):
        return hash((self.name, tuple(self.args)))

    def gen_sub(self, this: BaseType, typevars, cache=None) -> "Self":
        cache = cache or {}
        if me := cache.get(self):
            return me
        if len(self.args) == 0:
            return self
        res = object.__new__(self.__class__)  # todo: ugly... should make a clone()
        cache[self] = res
        if isinstance(this, TypeOperator):
            vardict = dict(zip(typevars.keys(), this.args))
        else:
            vardict = typevars
        for k in dataclasses.fields(self):
            setattr(res, k.name, getattr(self, k.name))
        res.args = [arg.resolve().gen_sub(this, vardict, cache) for arg in self.args]
        res.methods = {k: v.gen_sub(this, vardict, cache) for k, v in self.methods.items()}
        res.parents = [p.gen_sub(this, vardict, cache) for p in self.parents]
        res.is_protocol = self.is_protocol
        return res

    def to_list(self) -> List["BaseType"]:
        return [self, *self.args]

@dataclass
class ModuleType(TypeOperator):
    pass


class FunctionType(TypeOperator):
    def __init__(self, args: List[BaseType], ret: BaseType):
        super().__init__([ret, *args])

    @property
    def parameters(self):
        return self.args[1:]

    @property
    def return_type(self):
        return self.args[0]

    @return_type.setter
    def return_type(self, value):
        self.args[0] = value

    def __str__(self):
        ret, *args = map(str, self.args)
        if self.variadic:
            args.append(f"*args")
        if args:
            args = f"({', '.join(args)})"
        else:
            args = "()"
        return f"{args} -> {ret}"

    def remove_self(self):
        res = FunctionType(self.parameters[1:], self.return_type)
        res.variadic = self.variadic
        res.optional_at = self.optional_at - 1 if self.optional_at is not None else None
        return res


class CppType(TypeOperator):
    def __init__(self, name: str):
        super().__init__([name], name)

    def __str__(self):
        return self.name


class Union(TypeOperator):
    def __init__(self, left: BaseType, right: BaseType):
        super().__init__([left, right], "Union")


class TypeType(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "Type")

    @property
    def type_object(self) -> BaseType:
        return self.args[0]

    @type_object.setter
    def type_object(self, value: BaseType):
        self.args[0] = value


TY_TYPE = TypeOperator.make_type("type")
TY_INT = TypeOperator.make_type("int")
TY_FLOAT = TypeOperator.make_type("float")
TY_STR = TypeOperator.make_type("str")
TY_BYTES = TypeOperator.make_type("bytes")
TY_BOOL = TypeOperator.make_type("bool")
TY_COMPLEX = TypeOperator.make_type("complex")
TY_NONE = TypeOperator.make_type("NoneType")
#TY_MODULE = TypeOperator([], "module")
TY_VARARG = TypeOperator.make_type("vararg")
TY_SELF = TypeOperator.make_type("Self")
TY_SELF.gen_sub = lambda this, typevars, _: this


class PyList(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "list")

    @property
    def element_type(self):
        return self.args[0]


class PySet(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "set")

    @property
    def element_type(self):
        return self.args[0]


class PyDict(TypeOperator):
    def __init__(self, key: BaseType, value: BaseType):
        super().__init__([key, value], "dict")

    @property
    def key_type(self):
        return self.args[0]

    @property
    def value_type(self):
        return self.args[1]


class PyIterator(TypeOperator):
    def __init__(self, arg: BaseType):
        super().__init__([arg], "iter")

    @property
    def element_type(self):
        return self.args[0]


class TupleType(TypeOperator):
    def __init__(self, *args: List[BaseType]):
        super().__init__(args, "tuple")


class PromiseKind(Enum):
    TASK = 0
    JOIN = 1
    FUTURE = 2
    FORKED = 3
    GENERATOR = 4


class Promise(TypeOperator, ABC):
    def __init__(self, ret: BaseType, kind: PromiseKind):
        super().__init__([ret, MagicType(kind)])

    @property
    def return_type(self) -> BaseType:
        return self.args[0]

    @property
    def kind(self) -> PromiseKind:
        return self.args[1].val

    @kind.setter
    def kind(self, value: PromiseKind):
        if value == PromiseKind.GENERATOR:
            self.methods["__iter__"] = FunctionType([], self)
            self.methods["__next__"] = FunctionType([], self.return_type)
        self.args[1].val = value

    def __str__(self):
        return f"{self.kind.name.lower()}<{self.return_type}>"

    def get_parents(self) -> List["BaseType"]:
        if self.kind == PromiseKind.GENERATOR:
            return [*super().get_parents()]
        return super().get_parents()


class Forked(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.FORKED)


class Task(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.TASK)


class Future(Promise):
    """Only use this for type specs"""

    def __init__(self, ret: BaseType):
        super().__init__(ret, PromiseKind.FUTURE)


class UserType(TypeOperator):
    def __init__(self, name: str):
        super().__init__([], name=name, is_reference=True)

    def unify_internal(self, other: "BaseType"):
        if type(self) != type(other):
            raise IncompatibleTypesError()
